/*
The C++ fPoint class.
Copyright (C) 1999 E. Robert Tisdale

This file is part of a C++ class library used to help convert algorithms
from floating-point arithmetic to fixed-point arithmetic.
This library is free software which you can redistribute and/or modify
under the terms of the GNU Library General Public License
as published by the Free Software Foundation;
either version 2, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU Library General Public License
along with this library.  If not, write to the Free Software Foundation,
Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Written by E. Robert Tisdale
*/

#include<fPoint.h>

//const fPoint::ModeType fPoint::Mode = fPoint::floatingPoint;
  const fPoint::ModeType fPoint::Mode = fPoint::fixedPoint;

int
main() {
  // A binary two's complement fixed point number
  // \begin{equation}
  //	x = m2^{e}
  // \end{equation}
  // where the exponent $e$ is an integer, the mantissa
  // \begin{equation}
  //	m = -m_{0} + \sum_{k=1}^{p} m_{k}2^{-k}
  // \end{equation}
  // where the mantissa bits $m_{k} \in \{0, 1\} and
  // the least significant bit is in position $p \geq 0$
  // so that $-1 \leq m \leq +1 - 2^{-p}$
  // and $-2^{e} \leq x \leq +2^{e} - 2^{e-p}.

  // These flag bits control saturation and rounding.
  typedef	fPoint::Specification::flagbits flagbits;
  flagbits	ZeroBits = fPoint::Specification::ZeroBits;
  flagbits	Saturate = fPoint::Specification::Saturate;
  flagbits	Roundoff = fPoint::Specification::Roundoff;

  const
  int		exponent = 3;			// -8.0 <= x <= 8.0 - 0.5
  const
  unsigned int	position = 4;			// least significant bit
  const
  flagbits	controls = Saturate|Roundoff;	// round to nearest
  const
  char* const	notation = "this is a test";	// description
  
  fPoint::Specification
		s(exponent, position, controls, notation);
  cout << s;
					/*
Notation = this is a test
Exponent = 3
Position = 4
Controls = Roundoff
 Overflows = 0
Underflows = 0
					*/
  fPoint	a(s, -5.2777);
  cout << a << " = a(s, -5.2777)" << endl;
					/*
-5.5 = a(s, -5.2777)
					*/
  fPoint::DefaultSpecifier = &s;
  fPoint	b(-5.2777);
  cout << b << " = b(-5.2777)" << endl;
					/*
-5.5 = b(-5.2777)
					*/
  fPoint	x(s);
  cout << x << " = x(s)" << endl;
					/*
   NaN = x(s)
					*/
  x = a/b;
  cout << x << " = x = a/b" << endl;
					/*
1 = x = a/b
					*/
  fPoint	y(s, a + b);
  cout << y << " = y(s, a + b)" << endl;
					/*
5 = y(s, a + b)
					*/
  fPoint	z(s, x + y);
  cout << z << " = z(s, x + y)" << endl;
					/*
6 = z(s, x + y)
					*/
  z = fPoint(s, a/b) + y;
  cout << z << " = z = fPoint(s, a/b) + y" << endl;
					/*
6 = z = fPoint(s, a/b) + y
					*/
  z = (x = a/b) + y;
  cout << z << " = z = (x = a/b) + y" << endl;
					/*
6 = z = (x = a/b) + y
					*/
  cout << s;
					/*
Notation = this is a test
Exponent = 3
Position = 4
Controls = Roundoff
 Overflows = 1
Underflows = 0
					*/
  return 0;
  }
