#ifndef _fPoint_h
#define _fPoint_h 1
/*
The C++ fPoint class.
Copyright (C) 1999 E. Robert Tisdale

This file is part of a C++ class library used to help convert algorithms
from floating-point arithmetic to fixed-point arithmetic.
This library is free software which you can redistribute and/or modify
under the terms of the GNU Library General Public License
as published by the Free Software Foundation;
either version 2, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU Library General Public License
along with this library.  If not, write to the Free Software Foundation,
Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Written by E. Robert Tisdale
*/

#include<cmath>
#include<limits>
#include<climits>
#include<iomanip>

class fPoint {				// floating or fixed Point Arithmetic
public:
  enum ModeType {floatingPoint, fixedPoint};
  static
  const
  ModeType	Mode;
  class Specification;			// forward declaration
  class Statistics {
  private:
    // Representation
    unsigned
    int		Samples;		// number of coersions
    int		Minimum;		// minimum exponent
    int		Maximum;		// maximum exponent
    double	Average;		// average exponent
    double     Variance;		// deviation squared
    unsigned
    int	     Underflows;		// underflows
    unsigned
    int	      Overflows;		//  overflows
  public:
    // Constructors
		Statistics(void): Samples(0),
    Minimum(INT_MAX), Maximum(INT_MIN),
    Average((double)0), Variance((double)0),
    Underflows(0), Overflows(0) { }
    // Functions
    Statistics&	update(const double& x);// update statistics
    
    unsigned
    int		samples(void) const	{ return Samples; }
    int		minimum(void) const	{ return Minimum; }
    int		maximum(void) const	{ return Maximum; }
    double	average(void) const	{ return Average; }
    double     variance(void) const	{ return Variance; }
    unsigned
    int	     underflows(void) const	{ return Underflows; }
    unsigned
    int&     underflows(void)		{ return Underflows; }
    unsigned
    int	      overflows(void) const	{ return  Overflows; }
    unsigned
    int&      overflows(void)		{ return  Overflows; }
    // Friends
    friend class fPoint::Specification;
    };

  class	Specification: public Statistics {
  public:
    enum FlagBits {
      ZeroBits = 0,
      Saturate = 1 << 0,
      Roundoff = 1 << 1,
      Signless = 1 << 2,
      SignMagnitude = 1 << 3 };
    typedef unsigned int flagbits;
    static
    int		DefaultExponent;
    static
    unsigned
    int		DefaultPosition;
    static
    flagbits	DefaultControls;
    static
    const
    char*	DefaultNotation;
    static
    const int	min_exponent = numeric_limits<double>::min_exponent;
    static
    const int	max_exponent = numeric_limits<double>::max_exponent;
    static
    const int	digits = numeric_limits<double>::digits;
  private:
    // Representation
    const
    int		Exponent;		// fixed-point range
    const
    unsigned
    int		Position;		// fixed-point precision
    const
    flagbits	Controls;		// control rounding and overflow
    const
    char* const	Notation;		// notation string
  public:
    // Constructors
    explicit	Specification(int e = DefaultExponent,
		     unsigned int p = DefaultPosition,
			 flagbits c = DefaultControls,
		      const char* n = DefaultNotation);
		Specification(const Specification& s): Statistics(),
      Exponent(s.Exponent), Position(s.Position),
      Controls(s.Controls), Notation(s.Notation) { }
    // Functions
    int		exponent(void) const	{ return Exponent; }
    unsigned
    int		position(void) const	{ return Position; }
    flagbits	controls(void) const	{ return Controls; }
    const
    char* const	notation(void) const	{ return Notation; }
    bool	saturate(void) const	{ return Controls&Saturate; }
    bool	roundoff(void) const	{ return Controls&Roundoff; }
    bool	signless(void) const	{ return Controls&Signless; }
    bool   signmagnitude(void) const	{ return Controls&SignMagnitude; }
    double	roundoff(const double& x) const;
    double	  coerce(const double& x) const;
    };
  static const
  Specification  InitialSpecifier;
  static const
  Specification* DefaultSpecifier;

  class Result {
  private:
    // Representation
    double	Floating;
  public:
    // Constructors
		Result(const double& x = (double)0): Floating(x) { }
		Result(const fPoint& x);
    // Functions
    double	floating(void) const	{ return Floating; }
    // Operators
    Result&	operator  =(const double& x)	{ Floating = x; return *this; }
    Result&	operator  =(const fPoint& x);
    // No other operations are permitted on type Result.
    };
private:
  // Representation
  double		Floating;
  const
  Specification&	Specifier;
public:
  // Constructors
		fPoint(void);
  explicit	fPoint(const double& x);
  explicit	fPoint(const Result& x);
  explicit	fPoint(const Specification& s);
  explicit	fPoint(const Specification& s, const double& x);
  explicit	fPoint(const Specification& s, const Result& x);
  explicit	fPoint(const Specification& s, const fPoint& x);
  // Functions
  double	floating(void) const	{ return Floating; }
  const
  Specification&
	       specifier(void) const	{ return Specifier; }
  Statistics& statistics(void) const	{ return (Statistics&)Specifier; }
  fPoint&	  assign(const double& x);
  // Operators
  // unary minus
  Result	operator -(void) const	{ return -floating(); }
  // unary plus
  const
  fPoint&	operator +(void) const	{ return *this; }
  // cast (type conversion)
		operator const Specification& () const { return specifier(); }
		operator Statistics& () const { return statistics(); }
  // multiply
  Result	operator *(const double& x) const { return floating()*x; }
  Result	operator *(const fPoint& x) const {
    return floating()*x.floating(); }
  //   divide
  Result	operator /(const double& x) const { return floating()/x; }
  Result	operator /(const fPoint& x) const {
    return floating()/x.floating(); }
  //      add
  Result	operator +(const double& x) const { return floating() + x; }
  Result	operator +(const fPoint& x) const {
    return floating() + x.floating(); }
  // subtract
  Result	operator -(const double& x) const { return floating() - x; }
  Result	operator -(const fPoint& x) const {
    return floating() - x.floating(); }

  //  left
  Result	operator <<(int n) const { return ldexp(floating(), +n); }
  //  right
  Result	operator >>(int n) const { return ldexp(floating(), -n); }
  //    less than
  bool		operator < (const double& x) const { return floating() < x; }
  bool		operator < (const fPoint& x) const {
    return floating() < x.floating(); }
  //    less than or equal
  bool		operator <=(const double& x) const { return floating() <= x; }
  bool		operator <=(const fPoint& x) const {
    return floating() <= x.floating(); }
  // greater than
  bool		operator > (const double& x) const { return floating() > x; }
  bool		operator > (const fPoint& x) const {
    return floating() > x.floating(); }
  // greater than or equal
  bool		operator >=(const double& x) const { return floating() >= x; }
  bool		operator >=(const fPoint& x) const {
    return floating() >= x.floating(); }
  //     equal
  bool		operator ==(const double& x) const { return floating() == x; }
  bool		operator ==(const fPoint& x) const {
    return floating() == x.floating(); }
  // not equal
  bool		operator !=(const double& x) const { return floating() != x; }
  bool		operator !=(const fPoint& x) const {
    return floating() != x.floating(); }
  // simple assignment
  fPoint&	operator  =(const double& x) { return assign(x); }
  fPoint&	operator  =(const Result& x) {
    return operator  =(x.floating()); }
  fPoint&	operator  =(const fPoint& x) {
    return operator  =(x.floating()); }
  // multiply and assign
  fPoint&	operator *=(const double& x) {
    return operator  =(floating()*x); }
  fPoint&	operator *=(const fPoint& x) {
    return operator  =(floating()*x.floating()); }
  //   divide and assign
  fPoint&	operator /=(const double& x) {
    return operator  =(floating()/x); }
  fPoint&	operator /=(const fPoint& x) {
    return operator  =(floating()/x.floating()); }
  //      add and assign
  fPoint&	operator +=(const double& x) {
    return operator  =(floating() + x); }
  fPoint&	operator +=(const fPoint& x) {
    return operator  =(floating() + x.floating()); }
  // subtract and assign
  fPoint&	operator -=(const double& x) {
    return operator  =(floating() - x); }
  fPoint&	operator -=(const fPoint& x) {
    return operator  =(floating() - x.floating()); }
  //  left  and assign
  fPoint&	operator <<=(int n) {
    return operator  =(ldexp(floating(), +n)); }
  //  right and assign
  fPoint&	operator >>=(int n) {
    return operator  =(ldexp(floating(), -n)); }
  };

inline	fPoint::Result::Result(const fPoint& x): Floating(x.floating()) { }
inline
fPoint::Result&	fPoint::Result::operator  =(const fPoint& x) {
  Floating = x.floating(); return *this; }

// Constructors
inline	fPoint::fPoint(void):
  Floating(numeric_limits<double>::quiet_NaN()),
    Specifier(*DefaultSpecifier) { }
inline	fPoint::fPoint(const double& x):
    Floating(x), Specifier(*DefaultSpecifier) { assign(x); }
inline	fPoint::fPoint(const Result& x):
    Floating(x.floating()), Specifier(*DefaultSpecifier) { assign(x.floating()); }
inline	fPoint::fPoint(const Specification& s):
  Floating(numeric_limits<double>::quiet_NaN()), Specifier(s) { }
inline	fPoint::fPoint(const Specification& s, const double& x):
    Floating(x), Specifier(s) { assign(x); }
inline	fPoint::fPoint(const Specification& s, const Result& x):
    Floating(x.floating()), Specifier(s) { assign(x.floating()); }
inline	fPoint::fPoint(const Specification& s, const fPoint& x):
    Floating(x.floating()), Specifier(s) { assign(x.floating()); }

// Functions
inline
fPoint::Result	 abs(const fPoint& x) { return  ::fabs(x.floating()); }
inline
fPoint::Result	fabs(const fPoint& x) { return  ::fabs(x.floating()); }
inline
fPoint::Result	ceil(const fPoint& x) { return  ::ceil(x.floating()); }
inline
fPoint::Result floor(const fPoint& x) { return ::floor(x.floating()); }
inline
fPoint::Result	sqrt(const fPoint& x) { return  ::sqrt(x.floating()); }
inline
fPoint::Result	 pow(const fPoint& x, const double& e) {
  return ::pow(x.floating(), e); }
inline
fPoint::Result	 pow(const fPoint& x, const fPoint& e) {
  return ::pow(x.floating(), e.floating()); }
inline
fPoint::Result	 pow(const double& x, const fPoint& e) {
  return ::pow(x, e.floating()); }
inline
fPoint::Result	 pow(const fPoint& x, int i) { return ::pow(x.floating(), i); }
inline
fPoint::Result	 cos(const fPoint& x) { return   ::cos(x.floating()); }
inline
fPoint::Result	 sin(const fPoint& x) { return   ::sin(x.floating()); }
inline
fPoint::Result	 tan(const fPoint& x) { return   ::tan(x.floating()); }
inline
fPoint::Result	acos(const fPoint& x) { return  ::acos(x.floating()); }
inline
fPoint::Result	asin(const fPoint& x) { return  ::asin(x.floating()); }
inline
fPoint::Result	atan(const fPoint& x) { return  ::atan(x.floating()); }
inline
fPoint::Result atan2(const fPoint& x, const double& y) {
  return ::atan2(x.floating(), y); }
inline
fPoint::Result atan2(const fPoint& x, const fPoint& y) {
  return ::atan2(x.floating(), y.floating()); }
inline
fPoint::Result atan2(const double& x, const fPoint& y) {
  return ::atan2(x, y.floating()); }
inline
fPoint::Result	cosh(const fPoint& x) { return  ::cosh(x.floating()); }
inline
fPoint::Result	sinh(const fPoint& x) { return  ::sinh(x.floating()); }
inline
fPoint::Result	tanh(const fPoint& x) { return  ::tanh(x.floating()); }
inline
fPoint::Result	 exp(const fPoint& x) { return   ::exp(x.floating()); }
inline
fPoint::Result	 log(const fPoint& x) { return   ::log(x.floating()); }
inline
fPoint::Result log10(const fPoint& x) { return ::log10(x.floating()); }

// Operators
// multiply
inline
fPoint::Result	operator *(const double& x, const fPoint& y) {
  return x*y.floating(); }
//   divide
inline
fPoint::Result	operator /(const double& x, const fPoint& y) {
  return x/y.floating(); }
//      add
inline
fPoint::Result	operator +(const double& x, const fPoint& y) {
  return x + y.floating(); }
// subtract
inline
fPoint::Result	operator -(const double& x, const fPoint& y) {
  return x - y.floating(); }
//  left (output)
ostream&	operator <<(ostream& os, const fPoint::Statistics& s);
ostream&	operator <<(ostream& os, const fPoint::Specification& s);
ostream&	operator <<(ostream& os, const fPoint& x);
//  right (input)
istream&	operator <<(istream& is, const fPoint::Statistics& s);
istream&	operator <<(istream& is, const fPoint::Specification& s);
istream&	operator <<(istream& is,       fPoint& x);

#endif /* _doublefPoint_h */

