/*
The C++ fPoint class.
Copyright (C) 1999 E. Robert Tisdale

This file is part of a C++ class library used to help convert algorithms
from floating-point arithmetic to fixed-point arithmetic.
This library is free software which you can redistribute and/or modify
under the terms of the GNU Library General Public License
as published by the Free Software Foundation;
either version 2, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU Library General Public License
along with this library.  If not, write to the Free Software Foundation,
Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Written by E. Robert Tisdale
*/

#include<fPoint.h>
#include<string.h>

extern "C" {
  int finite(double);
  int  ilogb(double);
  };

int	ilg(const double& x) {
  return 1 + ilogb(x); }		// overflows unless x is finite
//int		e;			// x = f*2^{e}
//double	f = frexp(x, &e);	// 0.5 <= |f| <  1.0
//return e; }

// Functions
fPoint::Statistics&
  fPoint::Statistics::update(const double& x) {	// update statistics
  if ((double)0 == x) {
    ++underflows();
    }
  else					// x is not zero
  if (finite(x)) {			// x is finite
    ++Samples;
    int		e = ilg(x);
    if (e < Minimum) Minimum = e;
    if (e > Maximum) Maximum = e;
    Average  += (e - Average)/Samples;
    Variance += (x*x - Variance)/Samples;
    }
  else					// x is infinite or NaN
    ++overflows();
  return *this; }

// Operators
//  left (output)
ostream&	operator <<(ostream& os, const fPoint::Statistics& s) {
  os << " Overflows = "			<<  s.overflows()	<< endl;
  os << "Underflows = "			<< s.underflows()	<< endl;
  if (fPoint::Mode == fPoint::floatingPoint) {
    const
    int		width = os.width(0);
    os << "Samples = "			<<    s.samples()	<< endl;
    os << "Exponent: minimum = "	<<    s.minimum()
       << ", average = " << setw(width)	<<    s.average()
       << ", maximum = "		<<    s.maximum()	<< endl;
    os << "Variance = " << setw(width)	<<   s.variance()	<< endl;
    }
  return os; }
//  right (input)
istream&	operator >>(istream& is, const fPoint::Statistics& s) {
  // Don't change s.  Just read past the statistics in the input stream.
  const
  char		term = '\n';
  const
  int		n = 1024;
  char		p[n];
  if (is.getline(p, n, term) && strncmp(" Overflows = ", p, 13))
    if (is.getline(p, n, term) && strncmp("Underflows = ", p, 13))
      if (fPoint::Mode == fPoint::floatingPoint) {
	if (is.getline(p, n, term) && strncmp("Samples = ", p, 10))
	  if (is.getline(p, n, term) && strncmp("Exponent: minimum = ", p, 20))
	    if (is.getline(p, n, term) && strncmp("Variance = ", p, 11))
	      if (is.getline(p, n, term) && strncmp("Variance = ", p, 11));
	}
  return is; }

typedef fPoint::Specification::flagbits flagbits;
flagbits	ZeroBits = fPoint::Specification::ZeroBits;
int		fPoint::Specification::DefaultExponent =  0;
unsigned int	fPoint::Specification::DefaultPosition = 15;
flagbits	fPoint::Specification::DefaultControls = ZeroBits;
const char*	fPoint::Specification::DefaultNotation = "";

// Constructors
		fPoint::Specification::Specification(
    int e, unsigned int p, flagbits c, const char* n):
  Statistics(), Exponent(e), Position(p), Controls(c), Notation(n) {
  if ((unsigned int)digits < Position)
    cerr << "The floating-point type double does not have enough digits\n"
	 << "to represent all fixed-point numbers specified by\n"
	 << "Notation = " << Notation << endl;
  if (max_exponent < Exponent)
    cerr << "The floating-point type double has no representation\n"
	 << "large enough for all fixed-point numbers specified by\n"
	 << "Notation = " << Notation << endl;
  if (Exponent - (int)Position < min_exponent - digits)
    cerr << "The floating-point type double has no representation\n"
	 << "small enough for all fixed-point numbers specified by\n"
	 << "Notation = " << Notation << endl;
  };
// Functions
double		fPoint::Specification::roundoff(const double& x) const {
  return ldexp(floor(ldexp(x, (int)position() - exponent())
	   + 0.5*roundoff()), exponent() - (int)position()); }

double		fPoint::Specification::coerce(const double& x) const {
  double	t = x;			// return value
  if ((double)0 != t) {
    if (finite(t)) {			// t is a finite number
      Statistics&
		statistics = (Statistics&)(*this);
      const
      double	modulus = ldexp((double)1, +exponent());// 2^{exponent()}
      const
      double	epsilon = ldexp((double)1, -position());// 2^{position()}
      const
      double	minimum = -modulus;
      const
      double	maximum = +modulus*((double)1 - epsilon);

      int	e = ilg(t);
      if (exponent() + digits < e) {
	++statistics.overflows();
	t = (saturate())? (t < (double)0)? minimum: maximum: (double)0;
	}
      else {				// e <= exponent() + digits
	if (exponent() - ((int)position() + (int)roundoff()) < e) {
	  // First, check for overflow.
	  bool	overflow = false;
	  if (exponent() < e && minimum != t) {
	    overflow = true;
	    if (saturate()) {		//  overflow saturates
	      t = (t < (double)0)? minimum: maximum;
	      }
	    else {			//  overflow wraps around
	      double	s;
	      t = ldexp(modf(ldexp(t, -exponent()), &s), +exponent());
	      if (fmod(s, (double)2)) {	// s is odd
		if (t < (double)0)	// t is negative
		  t += modulus;
		else			// t is positive or zero
		  t -= modulus;
		}
	      }
	    }
	  // Then roundoff or truncate.
	  if (t < (double)0) {		// t is negative
	    if (minimum < t) {
	      int	e = ilg(t);
	      t = roundoff(t + ldexp((double)1, e)) - ldexp((double)1, e);
	      }
	    }
	  else
	  if ((double)0 < t) {		// t is positive
	    t = roundoff(t);
	    if (maximum < t) {
	      overflow = true;
	      t = (saturate())? maximum: minimum;
	      }
	    }
	  if (overflow)
	    ++statistics.overflows();
	  }
	else {	// e <= exponent() - ((int)position() + (int)roundoff())
	  if (t < (double)0 && !roundoff()) {
	    t = -modulus*epsilon;
	    }
	  else {
	    ++statistics.underflows();
	    t = (double)0;
	    }
	  }
	}
      }
    else {				// t is infinite or NaN
      cerr << t
	   << " = fPoint::Specification::coerce(const double&) const" << endl;
      }
    }
  return t;
  }

// Operators
//  left (output)
ostream&	operator <<(ostream& os, const fPoint::Specification& s) {
  os << "Notation = " << s.notation() << endl;
  os << "Exponent = " << s.exponent() << endl
     << "Position = " << s.position() << endl
     << "Controls = ";
  bool	previous = false;
  if (s.saturate()) {
    if (previous) os << "|"; else previous = true;
    os << "Saturate";
    }
  if (s.roundoff()) {
    if (previous) os << "|"; else previous = true;
    os << "Roundoff";
    }
  if (s.signmagnitude()) {
    if (previous) os << "|"; else previous = true;
    os << "SignMagnitude";
    }
  if (!previous) os << "ZeroBits";
  os << endl;
  os << (fPoint::Statistics&)s;
  return os; }
//  right (input)
istream&	operator >>(istream& is, const fPoint::Specification& s) {
  // Don't change s. Just read past the specification in the input stream.
  const
  char		term = '\n';
  const
  int		n = 1024;
  char		p[n];
  if (is.getline(p, n) && strncmp("Notation = ", p, 11)) {
    do {
      if(!is.getline(p, n, term))
        return is;
      } while (!strncmp("Exponent = ", p, 11));
    if (is.getline(p, n, term) && strncmp("Position = ", p, 11))
      if (is.getline(p, n, term) && strncmp("Controls = ", p, 11))
	is >> (fPoint::Statistics&)s;
    }
  return is; }

const
fPoint::Specification
		fPoint::InitialSpecifier = fPoint::Specification();
const
fPoint::Specification*
		fPoint::DefaultSpecifier = &fPoint::InitialSpecifier;

// Functions
fPoint&		fPoint::assign(const double& x) {
  if (Mode == fixedPoint) {
    Floating = specifier().coerce(x);
    }
  else {
    Floating = x;
    statistics().update(x);
    }
  return *this; }

// Operators
//  left (output)
ostream&	operator <<(ostream& os, const fPoint& x) {
  return os << x.floating(); }
//  right (input)
istream&	operator >>(istream& is,       fPoint& x) {
  double	t;
  if (is >> t)
    x = t;
  return is; }

